# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import dataclasses
import logging
import re
import typing as t

import docspec

from pydoc_markdown.interfaces import Processor, Resolver

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class _ParamLine:
  """
  Helper data class for holding details of Sphinx arguments.
  """

  name: str
  docs: str
  type: t.Optional[str] = None


def generate_sections_markdown(lines, sections):
  for key, section in sections.items():
    if lines and lines[-1]:
      lines.append('')

    lines.extend(['**{}**:'.format(key), ''])  # add an extra line because of markdown syntax
    lines.extend(section)


@dataclasses.dataclass
class SphinxProcessor(Processor):
  """
  This processor parses ReST/Sphinx-style function documentation and converts it into
  Markdown syntax.

  Example:

  ```
  :param arg1: This is the first argument.
  :raise ValueError: If *arg1* is a bad value.
  :return: An `int` that represents an interesting value.
  ```

  Renders as:

  :param arg1: This is the first argument.
  :raise ValueError: If *arg1* is a bad value.
  :return: An `int` that represents an interesting value.

  @doc:fmt:sphinx
  """

  def check_docstring_format(self, docstring: str) -> bool:
    return ':param' in docstring or ':return' in docstring or \
      ':raise' in docstring

  def process(self, modules: t.List[docspec.Module], resolver: t.Optional[Resolver]) -> None:
    docspec.visit(modules, self._process)

  def _process(self, node):
    if not node.docstring:
      return

    lines = []
    in_codeblock = False
    keyword = None
    components: t.Dict[str, t.List[str]] = {}
    parameters: t.List[_ParamLine] = []
    return_: t.Optional[_ParamLine] = None

    for line in node.docstring.split('\n'):
      if line.strip().startswith("```"):
        in_codeblock = not in_codeblock

      line_codeblock = line.startswith('    ')

      if not in_codeblock and not line_codeblock:
        line = line.strip()
        match = re.match(r'\s*:(arg|argument|param|parameter|type)\s+(\w+)\s*:(.*)?$', line)
        if match:
          keyword = 'Arguments'
          components.setdefault(keyword, [])
          param = match.group(2)
          text = match.group(3)
          text = text.strip()
          param_data = next((p for p in parameters if p.name == param), None)
          if match.group(1) == 'type':
            if param_data is None:
              param_data = _ParamLine(param, '', text)
              parameters.append(param_data)
            else:
              param_data.type = text
          else:
            if param_data is None:
              param_data = _ParamLine(param, text, None)
              parameters.append(param_data)
            else:
              param_data.docs = text
          continue

        match = re.match(r'\s*:(return|returns|rtype)\s*:(.*)?$', line)
        if match:
          keyword = 'Returns'
          components.setdefault('Returns', [])
          text = match.group(2)
          text = text.strip()
          if match.group(1) == 'rtype':
            if return_ is None:
              return_ = _ParamLine('return', '', text)
            else:
              return_.type = text
          else:
            if return_ is None:
              return_ = _ParamLine('return', text, None)
            else:
              return_.docs = text
          continue

        match = re.match('\\s*:(?:raises|raise)\\s+(\\w+)\\s*:(.*)?$', line)
        if match:
          keyword = 'Raises'
          exception = match.group(1)
          text = match.group(2)
          text = text.strip()

          component = components.setdefault(keyword, [])
          component.append('- `{}`: {}'.format(exception, text))
          continue

      if keyword is not None:
        components[keyword].append(line)
      else:
        lines.append(line)

    # Convert the parameters into actual markdown format.
    component = components['Arguments'] if parameters else []
    for param in parameters:
      if param.type:
        component.append('- `{}` (`{}`): {}'.format(param.name, param.type, param.docs))
      else:
        component.append('- `{}`: {}'.format(param.name, param.docs))

    # Convert the return data into markdown format.
    if return_:
      return_fmt = f'`{return_.type}`: {return_.docs}' if return_.type else return_.docs
      components['Returns'] = [return_fmt]

    generate_sections_markdown(lines, components)
    node.docstring = '\n'.join(lines)



# -*- coding: utf8 -*-

import dataclasses
import json
import logging
import os
import typing as t
import typing_extensions as te
from pathlib import Path

import databind.core.annotations as A
import docspec

from pydoc_markdown.interfaces import Renderer
from pydoc_markdown.contrib.renderers.markdown import MarkdownRenderer

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class CustomizedMarkdownRenderer(MarkdownRenderer):
  """We override some defaults in this subclass. """

  #: Disabled because Docusaurus supports this automatically.
  insert_header_anchors: bool = False

  #: Escape html in docstring, otherwise it could lead to invalid html.
  escape_html_in_docstring: bool = True

  #: Conforms to Docusaurus header format.
  render_module_header_template: str = (
    '---\n'
    'sidebar_label: {relative_module_name}\n'
    'title: {module_name}\n'
    '---\n\n'
  )


@dataclasses.dataclass
class DocusaurusRenderer(Renderer):
  """
  Produces Markdown files and a `sidebar.json` file for use in a [Docusaurus v2][1] websites.
  It creates files in a fixed layout that reflects the structure of the documented packages.
  The files will be rendered into the directory specified with the #docs_base_path option.

  Check out the complete [Docusaurus example on GitHub][2].

  [1]: https://v2.docusaurus.io/
  [2]: https://github.com/NiklasRosenstein/pydoc-markdown/tree/develop/examples/docusaurus

  ### Options
  """

  #: The #MarkdownRenderer configuration.
  markdown: te.Annotated[MarkdownRenderer, A.typeinfo(deserialize_as=CustomizedMarkdownRenderer)] = \
    dataclasses.field(default_factory=CustomizedMarkdownRenderer)

  #: The path where the docusaurus docs content is. Defaults "docs" folder.
  docs_base_path: str = 'docs'

  #: The output path inside the docs_base_path folder, used to ouput the
  #: module reference.
  relative_output_path: str = 'reference'

  #: The sidebar path inside the docs_base_path folder, used to ouput the
  #: sidebar for the module reference.
  relative_sidebar_path: str = 'sidebar.json'

  #: The top-level label in the sidebar. Default to 'Reference'. Can be set to null to
  #: remove the sidebar top-level all together. This option assumes that there is only one top-level module.
  sidebar_top_level_label: t.Optional[str] = 'Reference'

  #: The top-level module label in the sidebar. Default to null, meaning that the actual
  #: module name will be used. This option assumes that there is only one top-level module.
  sidebar_top_level_module_label: t.Optional[str] = None

  def render(self, modules: t.List[docspec.Module]) -> None:
    module_tree: t.Dict[str, t.Any] = {"children": {}, "edges": []}
    output_path = Path(self.docs_base_path) / self.relative_output_path
    for module in modules:
      filepath = output_path

      module_parts = module.name.split(".")
      if module.location.filename.endswith("__init__.py"):
        module_parts.append("__init__")

      relative_module_tree = module_tree
      intermediary_module = []

      for module_part in module_parts[:-1]:
        # update the module tree
        intermediary_module.append(module_part)
        intermediary_module_name = ".".join(intermediary_module)
        relative_module_tree["children"].setdefault(intermediary_module_name, {"children": {}, "edges": []})
        relative_module_tree = relative_module_tree["children"][intermediary_module_name]

        # descend to the file
        filepath = filepath / module_part

      # create intermediary missing directories and get the full path
      filepath.mkdir(parents=True, exist_ok=True)
      filepath = filepath / f"{module_parts[-1]}.md"

      with filepath.open('w') as fp:
        logger.info("Render file %s", filepath)
        self.markdown.render_to_stream([module], fp)

      # only update the relative module tree if the file is not empty
      relative_module_tree["edges"].append(
        os.path.splitext(str(filepath.relative_to(self.docs_base_path)))[0]
      )

    self._render_side_bar_config(module_tree)

  def _render_side_bar_config(self, module_tree: t.Dict[t.Text, t.Any]) -> None:
    """
    Render sidebar configuration in a JSON file. See Docusaurus sidebar structure:

    https://v2.docusaurus.io/docs/docs-introduction/#sidebar
    """
    sidebar = {
      "type": 'category',
      "label": self.sidebar_top_level_label,
    }
    self._build_sidebar_tree(sidebar, module_tree)

    if sidebar.get("items"):
      if self.sidebar_top_level_module_label:
        sidebar['items'][0]["label"] = self.sidebar_top_level_module_label

      if not self.sidebar_top_level_label:
        # it needs to be a dictionary, not a list; this assumes that
        # there is only one top-level module
        sidebar = sidebar['items'][0]

    sidebar_path = Path(self.docs_base_path) / self.relative_output_path / self.relative_sidebar_path
    with sidebar_path.open("w") as handle:
      logger.info("Render file %s", sidebar_path)
      json.dump(sidebar, handle, indent=2, sort_keys=True)

  def _build_sidebar_tree(self, sidebar: t.Dict[t.Text, t.Any], module_tree: t.Dict[t.Text, t.Any]) -> None:
    """
    Recursively build the sidebar tree, it follows Docusaurus sidebar structure:

    https://v2.docusaurus.io/docs/docs-introduction/#sidebar
    """
    sidebar["items"] = module_tree.get("edges", [])
    if os.name == 'nt':
      # Make generated configuration more portable across operating systems (see #129).
      sidebar["items"] = [x.replace('\\', '/') for x in sidebar["items"]]
    for child_name, child_tree in module_tree.get("children", {}).items():
      child = {
          "type": 'category',
          "label": child_name,
      }
      self._build_sidebar_tree(child, child_tree)
      sidebar["items"].append(child)

    def _sort_items(item: t.Union[t.Text, t.Dict[t.Text, t.Any]]) -> t.Tuple[int, t.Text]:
      """Sort sidebar items. Order follows:
        1. modules containing items come first
        2. alphanumeric order is applied
      """
      is_edge = int(isinstance(item, str))
      label = item if is_edge else item.get("label")
      return is_edge, label

    sidebar["items"] = sorted(sidebar["items"], key=_sort_items)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import dataclasses
import copy
import logging
import os
import subprocess
import typing as t
import typing_extensions as te

import databind.core.annotations as A
import docspec
import yaml

from pydoc_markdown.contrib.renderers.markdown import MarkdownRenderer
from pydoc_markdown.interfaces import Context, Renderer, Resolver, Server, Builder
from pydoc_markdown.util.pages import Page, Pages
from pydoc_markdown.util.knownfiles import KnownFiles
from typing import Dict, List, Optional

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class CustomizedMarkdownRenderer(MarkdownRenderer):
  """ We override some defaults in this subclass. """

  render_toc: bool = False


@dataclasses.dataclass
class MkdocsRenderer(Renderer, Server, Builder):
  """
  Produces Markdown files in a layout compatible with [MkDocs][0] and can be used with the
  Pydoc-Markdown `--server` option for a live-preview. The `--bootstrap mkdocs` option can
  be used to create a Pydoc-Markdown configuration file with the MkDocs template.

  Example configuration:

  ```yml
  renderer:
    type: mkdocs
    pages:
      - title: Home
        name: index
        source: README.md
      - title: API Documentation
        contents:
          - '*'
    mkdocs_config:
      mkdocs_config:
        site_name: My Project
        theme: readthedocs
  ```

  ### Options
  """

  #: The output directory for the generated Markdown files. Defaults to
  #: `build/docs`.
  output_directory: str = 'build/docs'

  #: Name of the content directory (inside the #output_directory). Defaults to "content".
  content_directory_name: str = 'content'

  #: Remove files generated in a previous pass by the Mkdocs renderer before rendering
  #: again. Defaults to `True`.
  clean_render: bool = True

  #: The pages to render into the output directory.
  # TODO (@NiklasRosenstein): Uh what exactly is this in new databind?
  pages: Pages[Page] = dataclasses.field(default_factory=Pages)

  #: Markdown renderer settings.
  # TODO (@NiklasRosenstein): Use fieldinfo(serialize_as)
  markdown: te.Annotated[MarkdownRenderer, A.typeinfo(deserialize_as=CustomizedMarkdownRenderer)] = \
    dataclasses.field(default_factory=CustomizedMarkdownRenderer)

  #: The name of the site. This will be carried into the `site_name` key
  #: of the #mkdocs_config.
  site_name: t.Optional[str] = None

  #: Arbitrary configuration values that will be rendered to an
  #: `mkdocs.yml` file.
  mkdocs_config: t.Optional[t.Dict[str, t.Any]] = dataclasses.field(default_factory=dict)

  #: Port for the Mkdocs server when using the `pydoc-markdown --server` option.
  #: Defaults to `8000`. Can be set with the `MKDOCS_PORT` environment variable:
  #:
  #: ```
  #: $ MKDOCS_PORT=8383 pydoc-markdown -so
  #: ```
  server_port: t.Optional[int] = None

  def __post_init__(self) -> None:
    self._context: t.Optional[Context] = None

  @property
  def content_dir(self) -> str:
    return os.path.join(self.output_directory, self.content_directory_name)

  def generate_mkdocs_nav(self, page_to_filename: Dict[Page, str]) -> Dict:
    def _generate(pages):
      result = []
      for page in pages:
        if page.children:
          result.append({page.title: _generate(page.children)})
        elif page.href:
          result.append({page.title: page.href})
        else:
          filename = os.path.relpath(page_to_filename[id(page)], self.content_dir)
          if os.name == 'nt':
            # Make generated configuration more portable across operating systems (see #129).
            filename = filename.replace('\\', '/')
          result.append({page.title: filename})
      return result
    return _generate(self.pages)

  def _get_addr(self) -> str:
    if self.server_port is None:
      port = int(os.getenv('MKDOCS_PORT', '8000'))
    else:
      port = self.server_port
    return 'localhost:{}'.format(port)

  # Renderer

  def render(self, modules: List[docspec.Module]) -> None:
    known_files = KnownFiles(self.output_directory)
    if self.clean_render:
      for file_ in known_files.load():
        try:
          os.remove(file_.name)
        except FileNotFoundError:
          pass

    page_to_filename = {}

    with known_files:
      for item in self.pages.iter_hierarchy():
        filename = item.filename(self.content_dir, '.md')
        page_to_filename[id(item.page)] = filename
        if not item.page.has_content():
          continue

        self.markdown.filename = filename
        item.page.render(filename, modules, self.markdown, context_directory=self._context.directory)
        known_files.append(filename)

      config = copy.deepcopy(self.mkdocs_config)
      if self.site_name:
        config['site_name'] = self.site_name
      if not config.get('site_name'):
        config['site_name'] = 'My Project'
      config['docs_dir'] = self.content_directory_name
      config['nav'] = self.generate_mkdocs_nav(page_to_filename)

      if self.mkdocs_config is not None:
        filename = os.path.join(self.output_directory, 'mkdocs.yml')
        logger.info('Rendering "%s"', filename)
        with known_files.open(filename, 'w') as fp:
          yaml.dump(config, fp)

  def get_resolver(self, modules: List[docspec.Module]) -> Optional[Resolver]:
    # TODO (@NiklasRosenstein): The resolver returned by the Markdown
    #   renderer does not implement linking across multiple pages.
    return self.markdown.get_resolver(modules)

  # Server

  def get_server_url(self) -> str:
    return 'http://{}'.format(self._get_addr())

  def start_server(self) -> subprocess.Popen:
    addr = self._get_addr()
    return subprocess.Popen(['mkdocs', 'serve', '-a', addr], cwd=self.output_directory)

  def reload_server(self, process: subprocess.Popen):
    # While MkDocs does support file watching and automatic reloading,
    # it appears to be a bit quirky. Let's just restart the whole process.
    process.terminate()

  # Builder

  def build(self, site_dir: str) -> None:
    command = ['mkdocs', 'build', '--clean', '--site-dir', site_dir]
    subprocess.check_call(command, cwd=self.output_directory)

  # PluginBase

  def init(self, context: Context) -> None:
    self._context = context
    self.markdown.init(context)
